import * as xlib from "xlib";
/** This is a specialized class used for commercial SAAS projects that have a free usage tier, and need to protect against abusive free-tier users.
 *
 * tracks a given resource, ensuring that multiple (free tier) userIds/ip's are not accessing it to get around free usage limits.
 *
 * once the abuse threshhold is passed, future requests to the resource are blacklisted.*/
export declare class UseLimitFraudCheck {
    options: {
        /** the maximum number of usages from seperate userIds to allow during a sample window.   example = 5 */
        maxValidUses: number;
        /** how long to track usages.   only usages in this interval are considered.  example = 5min. */
        sampleWindowMs: number | xlib.time.luxon.Duration;
        /** if max uses is reached, we will blacklist.   but delay by this long to make the blacklist reason less obvious.  example = 5min. */
        obfuscateBlacklistDelayMs: number | xlib.time.luxon.Duration;
        /** if this number of uses is reached while we are obfuscating (delay blacklisting) we will blacklist immediately.  example value=10 */
        blacklistImmediateUses: number;
        /** once blacklisted, how long you want the blacklist to take effect.  example value=1day */
        blacklistDuration: number | xlib.time.luxon.Duration;
    };
    constructor(options: {
        /** the maximum number of usages from seperate userIds to allow during a sample window.   example = 5 */
        maxValidUses: number;
        /** how long to track usages.   only usages in this interval are considered.  example = 5min. */
        sampleWindowMs: number | xlib.time.luxon.Duration;
        /** if max uses is reached, we will blacklist.   but delay by this long to make the blacklist reason less obvious.  example = 5min. */
        obfuscateBlacklistDelayMs: number | xlib.time.luxon.Duration;
        /** if this number of uses is reached while we are obfuscating (delay blacklisting) we will blacklist immediately.  example value=10 */
        blacklistImmediateUses: number;
        /** once blacklisted, how long you want the blacklist to take effect.  example value=1day */
        blacklistDuration: number | xlib.time.luxon.Duration;
    });
    /** @hidden
     *
     * ```Do not use.   exposed for debugging purposes only.```
     *
     * stores all resources and their associated tracking data.  key = resource
     *
     */
    _storage: Map<string, IResourceInfo>;
    /** private helper that cleans up an entry in our _storage Map. */
    private _doHousekeeping;
    checkUse(userId: string, resource: string): {
        /** true if we detect abusive behaviour from this ```userKey```'s access of the ```resource```,
         * regardless of if the accesss should be blocked */
        isAbuse: boolean;
        /** ```false``` if you should allow usage to the resource (such as non-fraud access, or fraud usage is detected but we are in our obfuscation grace period).
         *
         * An example would be to allow the abuse call, but penalize in cost.
         *
         * ```true``` means deny the usage (and inform the use as such.  EG HTTP 403 error)
         *  */
        preventAccess: boolean;
    };
}
/** if this account is a free account (unpaid registered or demo):  then track the request domain.  if more than 5 free accounts request the same domain in less than 5 minutes, blacklist the domain for all free users for 24hrs.   					 */
interface IResourceInfo {
    /**freebee requests to this domain in the last 5 minutes */
    requests: Map<string, Date>;
    /** we blacklist for 24hrs (GCP preemptive instances have a lifespan of less than that) */
    blacklistExpires?: xlib.time.luxon.DateTime;
    /** obfuscate blacklisting: once we decide to blacklist, we delay by 5 minutes or 10 accounts, whichever occurs first */
    blacklistStarts?: xlib.time.luxon.DateTime;
    /** callback used to cleanup after 5 minutes of the last request. */
    cleanupTimeoutHandle?: any;
}
export {};
//# sourceMappingURL=saas-abuse.d.ts.map